/// <reference types="node" />
import * as ByteBuffer from 'bytebuffer';
export interface ITransaction<AssetType = {}> {
    recipientId: string;
    senderId: string;
    amount: number;
    senderPublicKey: string;
    requesterPublicKey: string;
    timestamp: number;
    fee: number;
    asset: AssetType;
    type: number;
    id: string;
    signature: string;
    signSignature?: string;
    signatures?: string[];
}
/**
 * Base transaction class.
 */
export declare abstract class BaseTx<T = {}> {
    asset?: T;
    recipientId: string;
    amount: number;
    senderPublicKey: string;
    requesterPublicKey: string;
    timestamp: number;
    fee: number;
    protected abstract type: number;
    protected _signature: string;
    protected _secondSignature?: string;
    protected _id: string;
    private _addressSuffixLength;
    constructor(asset?: T);
    /**
     * Calculates bytes of tx.
     * @param {boolean} skipSignature=false true if you don't want to account signature
     * @param {boolean} skipSecondSign=false true if you don't want to account second signature
     * @returns {Buffer}
     */
    getBytes(skipSignature?: boolean, skipSecondSign?: boolean): Buffer;
    /**
     * Use the sign method within the wallet instance.
     * @internal
     */
    sign(signingPrivKey: string | {
        publicKey: string;
        privKey: string;
    }, signingSecondPrivKey?: string): ITransaction<T>;
    /**
     * Gets raw hash of current tx
     */
    getHash(): Buffer;
    set(key: 'type' | 'fee' | 'amount' | 'timestamp', value: number): this;
    set(key: 'senderPublicKey' | 'recipientId' | 'requesterPublicKey', value: string): this;
    set(key: 'asset', value: T): this;
    withRecipientId(recipientId: string): this;
    withAmount(amount: number): this;
    withSenderPublicKey(senderPublicKey: string): this;
    withRequesterPublicKey(senderPublicKey: string): this;
    withTimestamp(timestamp: number): this;
    withFees(fees: number): this;
    /**
     * Returns plain object representation of tx (if not signed error will be thrown)
     * Be aware that this method returns null senderId. You'll have to set to proper address yourself
     */
    toObj(): ITransaction<T>;
    /**
     * Generate signature from given private key.
     * @param {string} privKey
     * @returns {Buffer}
     */
    createSignature(privKey: string): any;
    /**
    * Set signature
    * @param {string | Buffer} signature
    */
    signature: string | Buffer;
    secondSignature: string | Buffer;
    signSignature: string | Buffer;
    id: string;
    /**
     * Set a different address suffix length (default is 1)
     */
    addressSuffixLength: number;
    /**
     * Calculates Tx id!
     * @returns {string}
     */
    calcId(): string;
    /**
     * Override to calculate asset bytes.
     * @param {boolean} skipSignature
     * @param {boolean} skipSecondSign
     */
    protected abstract getChildBytes(skipSignature: boolean, skipSecondSign: boolean): Buffer;
    /**
     * override this to allow asset and other fields creations.
     * for different tx types.
     */
    protected innerCreate(): void;
    /**
     * Utility to copy an hex string to a bytebuffer
     * @param {string} hex
     * @param {ByteBuffer} bb
     */
    static hexKeyInByteBuffer(hex: string, bb: ByteBuffer): void;
}
