"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var bignumber_js_1 = require("bignumber.js");
var ByteBuffer = require("bytebuffer");
var empty = require("is-empty");
var sodium_1 = require("../utils/sodium");
var bignumber_1 = require("../utils/bignumber");
var sha256_1 = require("../utils/sha256");
/**
 * Base transaction class.
 */
var BaseTx = /** @class */ (function () {
    // tslint:enable variable-name
    function BaseTx(asset) {
        this.asset = asset;
        this.requesterPublicKey = null;
        this._addressSuffixLength = 1;
    }
    /**
     * Calculates bytes of tx.
     * @param {boolean} skipSignature=false true if you don't want to account signature
     * @param {boolean} skipSecondSign=false true if you don't want to account second signature
     * @returns {Buffer}
     */
    BaseTx.prototype.getBytes = function (skipSignature, skipSecondSign) {
        if (skipSignature === void 0) { skipSignature = false; }
        if (skipSecondSign === void 0) { skipSecondSign = false; }
        var childBytes = this.getChildBytes(skipSignature, skipSecondSign);
        var assetSize = empty(childBytes) ? 0 : childBytes.length;
        var bb = new ByteBuffer(1 + 4 + 32 + 32 + 8 + 8 + 64 + 64 + assetSize, true);
        bb.writeByte(this.type);
        bb.writeInt(this.timestamp);
        BaseTx.hexKeyInByteBuffer(this.senderPublicKey, bb);
        if (!empty(this.requesterPublicKey)) {
            BaseTx.hexKeyInByteBuffer(this.requesterPublicKey, bb);
        }
        if (!empty(this.recipientId)) {
            var recipient = bignumber_1.bigNumberToBuffer(new bignumber_js_1.default(this.recipientId.slice(0, -this._addressSuffixLength)), { size: 8 });
            for (var i = 0; i < 8; i++) {
                bb.writeByte(recipient[i] || 0);
            }
        }
        else {
            for (var i = 0; i < 8; i++) {
                bb.writeByte(0);
            }
        }
        // tslint:disable-next-line no-string-literal
        bb['writeLong'](this.amount);
        if (assetSize > 0) {
            for (var i = 0; i < assetSize; i++) {
                bb.writeByte(childBytes[i]);
            }
        }
        if (!skipSignature && !empty(this._signature)) {
            BaseTx.hexKeyInByteBuffer(this._signature, bb);
        }
        if (!skipSecondSign && !empty(this._secondSignature)) {
            BaseTx.hexKeyInByteBuffer(this._secondSignature, bb);
        }
        bb.flip();
        // TODO: Check? this returns an array buffer which does not
        // inherit from buffer (according to ts types).
        return new Buffer(bb.toBuffer());
    };
    /**
     * Use the sign method within the wallet instance.
     * @internal
     */
    BaseTx.prototype.sign = function (signingPrivKey, signingSecondPrivKey) {
        var privKey;
        var wallet = null;
        if (typeof (signingPrivKey) === 'string') {
            privKey = signingPrivKey;
        }
        else if (signingPrivKey === null) {
            privKey = null;
        }
        else {
            privKey = signingPrivKey.privKey;
            wallet = signingPrivKey;
        }
        if (empty(this.type) && this.type !== 0) {
            throw new Error("Unknown transaction type " + this.type);
        }
        if (empty(this.senderPublicKey)) {
            if (wallet === null) {
                throw new Error('Sender Public Key is empty');
            }
            this.senderPublicKey = wallet.publicKey;
        }
        if ((empty(this.timestamp) && this.timestamp !== 0) || this.timestamp < 0) {
            throw new Error('Invalid timestamp provided');
        }
        if (empty(this.fee)) {
            throw new Error('No fees specified');
        }
        this.innerCreate();
        this._signature = this.createSignature(privKey).toString('hex');
        if (!empty(signingSecondPrivKey)) {
            this._secondSignature = this.createSignature(signingSecondPrivKey).toString('hex');
        }
        this._id = this.calcId();
        return this.toObj();
    };
    /**
     * Gets raw hash of current tx
     */
    BaseTx.prototype.getHash = function () {
        return sha256_1.toSha256(this.getBytes());
    };
    BaseTx.prototype.set = function (key, value) {
        this[key] = value;
        return this;
    };
    BaseTx.prototype.withRecipientId = function (recipientId) {
        return this.set('recipientId', recipientId);
    };
    BaseTx.prototype.withAmount = function (amount) {
        return this.set('amount', amount);
    };
    BaseTx.prototype.withSenderPublicKey = function (senderPublicKey) {
        return this.set('senderPublicKey', senderPublicKey);
    };
    BaseTx.prototype.withRequesterPublicKey = function (senderPublicKey) {
        return this.set('requesterPublicKey', senderPublicKey);
    };
    BaseTx.prototype.withTimestamp = function (timestamp) {
        return this.set('timestamp', timestamp);
    };
    BaseTx.prototype.withFees = function (fees) {
        return this.set('fee', fees);
    };
    /**
     * Returns plain object representation of tx (if not signed error will be thrown)
     * Be aware that this method returns null senderId. You'll have to set to proper address yourself
     */
    BaseTx.prototype.toObj = function () {
        if (empty(this._signature)) {
            throw new Error('Signature must be set before calling toObj');
        }
        if (empty(this._id)) {
            this._id = this.calcId();
        }
        // tslint:disable object-literal-sort-keys
        var toRet = {
            id: this._id,
            fee: this.fee,
            type: this.type,
            recipientId: this.recipientId,
            amount: this.amount,
            senderPublicKey: this.senderPublicKey,
            requesterPublicKey: this.requesterPublicKey,
            timestamp: this.timestamp,
            senderId: null,
            signature: this._signature,
            signSignature: this._secondSignature || undefined,
            asset: this.asset,
        };
        // tslint:enable object-literal-sort-keys
        if (empty(toRet.signSignature)) {
            delete toRet.signSignature;
        }
        return toRet;
    };
    /**
     * Generate signature from given private key.
     * @param {string} privKey
     * @returns {Buffer}
     */
    BaseTx.prototype.createSignature = function (privKey) {
        var hash = this.getHash();
        return sodium_1.api.crypto_sign_detached(hash, new Buffer(privKey, 'hex'));
    };
    Object.defineProperty(BaseTx.prototype, "signature", {
        get: function () {
            return this._signature;
        },
        /**
         * Set signature
         * @param {string | Buffer} signature
         */
        set: function (signature) {
            if (empty(signature)) {
                this._signature = null;
                return;
            }
            this._signature = (signature instanceof Buffer ? signature : new Buffer(signature, 'hex')).toString('hex');
            if (this._signature.length !== 128) {
                // Signature should have 64 bytes!
                throw new Error('Signature is not having a correct lengthness');
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BaseTx.prototype, "secondSignature", {
        get: function () {
            return this._secondSignature;
        },
        set: function (signature) {
            if (empty(signature)) {
                this._secondSignature = null;
                return;
            }
            if (empty(this._signature)) {
                throw new Error('Did you set signature first? Most likely you did not calculated 2nd Signature on correct bytes');
            }
            this._secondSignature = (signature instanceof Buffer ? signature : new Buffer(signature, 'hex')).toString('hex');
            if (this._secondSignature.length !== 128) {
                // Signature should have 64 bytes!
                throw new Error('Signature is not having a correct lengthness');
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BaseTx.prototype, "signSignature", {
        set: function (signature) {
            this.secondSignature = signature;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BaseTx.prototype, "id", {
        get: function () {
            if (empty(this._id)) {
                throw new Error('Call create first');
            }
            return this._id;
        },
        set: function (id) {
            this._id = id;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BaseTx.prototype, "addressSuffixLength", {
        /**
         * Set a different address suffix length (default is 1)
         */
        set: function (value) {
            this._addressSuffixLength = value;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Calculates Tx id!
     * @returns {string}
     */
    BaseTx.prototype.calcId = function () {
        var hash = this.getHash();
        var temp = new Buffer(8);
        for (var i = 0; i < 8; i++) {
            temp[i] = hash[7 - i];
        }
        return bignumber_1.bigNumberFromBuffer(temp).toString();
    };
    /**
     * override this to allow asset and other fields creations.
     * for different tx types.
     */
    // tslint:disable-next-line no-empty
    BaseTx.prototype.innerCreate = function () {
    };
    /**
     * Utility to copy an hex string to a bytebuffer
     * @param {string} hex
     * @param {ByteBuffer} bb
     */
    // tslint:disable-next-line member-ordering
    BaseTx.hexKeyInByteBuffer = function (hex, bb) {
        var buf = Buffer.from(hex, 'hex');
        // tslint:disable-next-line prefer-for-of
        for (var i = 0; i < buf.length; i++) {
            bb.writeByte(buf[i]);
        }
    };
    return BaseTx;
}());
exports.BaseTx = BaseTx;
