"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var BaseTx_1 = require("./trxTypes/BaseTx");
var sha256_1 = require("./utils/sha256");
var sodium_1 = require("./utils/sodium");
var txFactory_1 = require("./utils/txFactory");
// tslint:disable-next-line no-var-requires
/**
 * Generic Abstract Wallet.
 */
var GenericWallet = /** @class */ (function () {
    // tslint:enable variable-name
    function GenericWallet(secret) {
        this.createKeyPair(secret);
    }
    /**
     * Verifies message. Returns true or throws error if verification fails
     * @param {string | Buffer} message
     * @param {string | Buffer} signature
     * @param {string | Buffer} pubKey
     * @returns {true}
     */
    GenericWallet.verifyMessage = function (message, signature, pubKey) {
        if (typeof (message) === 'string') {
            message = new Buffer(message, 'utf8');
        }
        if (typeof (signature) === 'string') {
            signature = new Buffer(signature, 'hex');
        }
        if (typeof (pubKey) === 'string') {
            pubKey = new Buffer(pubKey, 'hex');
        }
        if (pubKey.length !== 32) {
            throw new Error('Invalid public Key');
        }
        var open = sodium_1.api.crypto_sign_open(signature, pubKey);
        if (!open) {
            throw new Error('Invalid signature. Cannot verify message');
        }
        if (open.toString('hex') !== message.toString('hex')) {
            throw new Error('Signature is valid but different message');
        }
        return true;
    };
    GenericWallet.prototype.signMessage = function (message) {
        if (typeof (message) === 'string') {
            message = new Buffer(message, 'utf8');
        }
        return sodium_1.api.crypto_sign(message, new Buffer(this.privKey, 'hex'));
    };
    /**
     * Signs a transaction
     * @param {ITransaction<T> | BaseTx<T>} tx
     * @param {GenericWallet} coSign optional cosigning wallet.
     * @returns {ITransaction<T>} signed transaction
     */
    GenericWallet.prototype.signTransaction = function (tx, coSign) {
        if (!(tx instanceof BaseTx_1.BaseTx)) {
            tx = txFactory_1.createTransactionFromOBJ(tx);
        }
        tx.signature = this.getSignatureOfTransaction(tx);
        if (coSign) {
            tx.signSignature = coSign.getSignatureOfTransaction(tx);
            tx.id = tx.calcId();
        }
        return __assign({}, tx.toObj(), { senderId: this.address });
    };
    /**
     * Returns signature of provided transaction without modifiying the original tx.
     * @return {string} signature buffer.
     */
    GenericWallet.prototype.getSignatureOfTransaction = function (tx) {
        if (!(tx instanceof BaseTx_1.BaseTx)) {
            tx = txFactory_1.createTransactionFromOBJ(tx);
        }
        tx.addressSuffixLength = this.addressOptions.suffixLength;
        if (!tx.signature) {
            var signedTx = tx.sign(this);
            return signedTx.signature;
        }
        return tx
            .createSignature(this.privKey)
            .toString('hex');
    };
    /**
     * Creates key pair from secret string
     * @param {string} secret
     */
    GenericWallet.prototype.createKeyPair = function (secret) {
        var hash = sha256_1.toSha256(secret);
        var keypair = sodium_1.api.crypto_sign_seed_keypair(hash);
        this._privKey = keypair.secretKey.toString('hex');
        this._publicKey = keypair.publicKey.toString('hex');
    };
    Object.defineProperty(GenericWallet.prototype, "privKey", {
        /**
         * @returns {string} privateKey
         */
        get: function () {
            return this._privKey;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GenericWallet.prototype, "publicKey", {
        /**
         * @returns {string} publicKey
         */
        get: function () {
            return this._publicKey;
        },
        enumerable: true,
        configurable: true
    });
    return GenericWallet;
}());
exports.GenericWallet = GenericWallet;
